//---------------------------------------------------------------------------
#include <stdio.h>  // printf
#include <stdlib.h> // atoi
#include <string.h>
//---------------------------------------------------------------------------
#include "kernel.h"
#include "tboolfnct.h"
#include "checks.h"
#include "buffers.h"
//---------------------------------------------------------------------------
MinimizeOptions * mo;
GeneralOptions * go;
TBuffers * buffers;
MinInfo info;
TBoolFnct * srcf, * minf;
TGauge gauge;
char * bc_file;
//---------------------------------------------------------------------------
int NewGauge( char * name, int max, int context )
{
  return 0;
};
//---------------------------------------------------------------------------
void CloseGauge()
{
};
//---------------------------------------------------------------------------
int GaugeStep( int val, int context )
{
  return 0;
};
//---------------------------------------------------------------------------
void help()
{
  fprintf( stderr, "syntax: BOOM [options] source [destination]\n" );
  fprintf( stderr, "Minimization Options:\n" );
  fprintf( stderr, "    -FRn  Define FC-Min:CD-search ratio n (0-100), 0 is default\n" );
  fprintf( stderr, "    -DFn  Define the FC-Min Depth Factor, 10 is default\n" );
  fprintf( stderr, "    -DDx  Define the FC-Min Depth Factor Direction\n" );
  fprintf( stderr, "             0   DF = 1:n (from -DFn)\n" );
  fprintf( stderr, "             1   DF = n:1 (from -DFn), default\n" );
  fprintf( stderr, "    -CMn  Define CD-search mutations ratio n (0-100)\n" );
  fprintf( stderr, "    -RMn  Define implicant reduction mutations ratio n (0-100)\n" );
  fprintf( stderr, "    -Ex   Select implicant expansion type:\n" );
  fprintf( stderr, "             0   Sequential search\n" );
  fprintf( stderr, "             1   Distributed multiple IE\n" );
  fprintf( stderr, "             2   Distributed exhaustive IE\n" );
  fprintf( stderr, "             3   Multiple IE (default)\n" );
  fprintf( stderr, "             4   Exhaustive IE\n" );
  fprintf( stderr, "    -CPx  Select the covering problem solution algorithm:\n" );
  fprintf( stderr, "             0   LCMC\n" );
  fprintf( stderr, "             1   Contributive selection (default)\n" );
  fprintf( stderr, "             2   Contributive removal\n" );
  fprintf( stderr, "             3   Exact cover\n" );
  fprintf( stderr, "    -Sxn  Define stopping criterion x of value n:\n" );
  fprintf( stderr, "             t   stop after n seconds\n" );
  fprintf( stderr, "             i   stop after n iterations (default is Si1)\n" );
  fprintf( stderr, "             n   stopping interval equal to n\n" );
  fprintf( stderr, "             q   stop when quality of solution meets n\n" );
  fprintf( stderr, "    -Qx   Define optimization criterion x:\n" );
  fprintf( stderr, "             t   terms\n" );
  fprintf( stderr, "             l   literals\n" );
  fprintf( stderr, "             o   output cost\n" );
  fprintf( stderr, "             b   literals+output cost (default)\n" );
  fprintf( stderr, "             g   number of gates\n" );
  fprintf( stderr, "             G   number of gate equivalents\n" );
  fprintf( stderr, "    -single   Single-output minimization\n" );
  fprintf( stderr, "    -endcov   Cover at the end only\n" );
  fprintf( stderr, "    -SD   Substitute output don't cares\n" );
  fprintf( stderr, "\n" );
  fprintf( stderr, "Special Functions (cannot be combined with minimization):\n" );
  fprintf( stderr, "    -info        Prints the information about the PLA function (for fr type only)\n" );
  fprintf( stderr, "    -Tcheck      Checks the input function for consistence\n" );
  fprintf( stderr, "    -Tsplit      Splits intersecting terms\n" );
  fprintf( stderr, "    -Tminterm    Splits function into minterms\n" );
}
//---------------------------------------------------------------------------
int main( int argc, char * argv[] )
{
  int i;
  char * inm, * onm;
  int op;
  // 0 = minimization
  // 1 = info
  // 2 = Tcheck
  // 3 = Tsplit
  // 4 = Tminterm

  gauge.NewGauge = NewGauge;
  gauge.CloseGauge = CloseGauge;
  gauge.GaugeStep = GaugeStep;

  mo = new MinimizeOptions;
  go = new GeneralOptions;

  InitOpts();

  inm = 0;
  onm = 0;
  i = 1;
  op = 0;
  srcf = 0;
  minf = 0;

  fprintf( stderr, "BOOM 2.2 beta for Win32 by Petr Fiser, (c) April 2004\n\n" );
  while ( i != argc )
  {
    if ( argv[i] [0] == '-' )
    {
      switch ( argv[i] [1] )
      {
        case 'F':
          switch ( argv[i] [2] )
          {
            case 'R':
              mo->CD_FC = atoi( & argv[i] [3] );
            break;
            default:
              fprintf( stderr, "Unknown option %c\n", argv[i] [1] );
              exit( 1 );
          }
        break;
        case 'D':
          switch ( argv[i] [2] )
          {
            case 'F':
              mo->FC_DF = atoi( & argv[i] [3] );
            break;
            case 'D':
              mo->FC_DF_dir = atoi( & argv[i] [3] );
            break;
            default:
              fprintf( stderr, "Unknown option %c\n", argv[i] [1] );
              exit( 1 );
          }
        break;
        case 'S':
          switch ( argv[i] [2] )
          {
            case 't':
              mo->STOP_time = 1;
              mo->STOP_maxtime = atof( & argv[i] [3] );
            break;
            case 'i':
              mo->STOP_iterations = 1;
              mo->STOP_maxiterations = atoi( & argv[i] [3] );
            break;
            case 'n':
              mo->STOP_improvement = 1;
              mo->STOP_improvementint = atoi( & argv[i] [3] );
            break;
            case 'q':
              mo->STOP_optimization = 1;
              mo->STOP_optimizationval = atoi( & argv[i] [3] );
            break;

            case 'D':
              mo->COV_substDC = 1;
            break;

            default:
              fprintf( stderr, "Unknown option %c\n", argv[i] [1] );
              exit( 1 );
          }
        break;
        case 'Q':
          switch ( argv[i] [2] )
          {
            case 't':
              mo->OPT_crit = 0;
            break;
            case 'l':
              mo->OPT_crit = 1;
            break;
            case 'o':
              mo->OPT_crit = 2;
            break;
            case 'b':
              mo->OPT_crit = 3;
            break;
            case 'g':
              mo->OPT_crit = 4;
            break;
            case 'G':
              mo->OPT_crit = 5;
            break;
            default:
              fprintf( stderr, "Unknown option %c\n", argv[i] [1] );
              exit( 1 );
          }
        break;
        case 'C':
          switch ( argv[i] [2] )
          {
            case 'M':
              mo->CD_mutations = atoi( & argv[i] [3] );
            break;
            case 'P':
              mo->COV_type = atoi( & argv[i] [3] );
            break;
            default:
              fprintf( stderr, "Unknown option %c\n", argv[i] [1] );
              exit( 1 );
          }
        break;

        case 'R':
          switch ( argv[i] [2] )
          {
            case 'M':
              mo->IR_mutations = atoi( & argv[i] [3] );
            break;
            default:
              fprintf( stderr, "Unknown option %c\n", argv[i] [1] );
              exit( 1 );
          }
        break;
        case 'E':
          mo->IE_type = atoi( & argv[i] [2] );
        break;
        case 'e':
          if ( strcmp( argv[i], "-endcov" ) == 0 )
          {
            mo->COV_end = 1;
          break;
          }
          else
          {
            fprintf( stderr, "Unknown parameter- %s\n", argv[i] );
            exit( 1 );
          }
        break;
        case 's':
          if ( strcmp( argv[i], "-single" ) == 0 )
          {
            mo->COV_somin = 1;
          break;
          }
        case 'i':
          if ( strcmp( argv[i], "-info" ) == 0 )
          {
            op = 1;
          break;
          }
        case 'T':
          if ( strcmp( argv[i], "-Tcheck" ) == 0 )
          {
            op = 2;
          break;
          }
          else if ( strcmp( argv[i], "-Tsplit" ) == 0 )
          {
            op = 3;
          break;
          }
          else if ( strcmp( argv[i], "-Tminterm" ) == 0 )
          {
            op = 4;
          break;
          }
          else
          {
            fprintf( stderr, "Unknown parameter %s\n", argv[i] );
          break;
          }

        default:
          fprintf( stderr, "Unknown parameter %s\n", argv[i] );
          exit( 1 );
      }
    }
    else if ( inm == 0 ) inm = argv[i];
    else
      onm = argv[i];
    i++;
  }
  if ( inm == 0 ) help();
  else
  {
    fprintf( stderr, "Input function: %s\n", inm );
    srcf = srcf->LoadFnct( inm );
    if ( srcf == 0 )
    {
      fprintf( stderr, "File not found\n" );
      return 1;
    }

    srcf->InitSourceFnct(); // InitDecompAss(srcf);
    if ( srcf->type == 1 ) fprintf( stderr, "Function type: fr\n" );
    else
      fprintf( stderr, "Function type: fd\n" );
    fprintf( stderr, "Input variables: %i\n", srcf->inpvars );
    fprintf( stderr, "Output variables: %i\n", srcf->outvars );
    fprintf( stderr, "Defined terms: %i\n", srcf->terms );
    fprintf( stderr, "\n" );
    switch ( op )
    {
      case 0:
        if ( mo->CD_FC > 100 )
        {
          fprintf( stderr, "CD-Search:FC-Min ratio must be between 0-100\n" );
          exit( 1 );
        }
        if ( mo->FC_DF_dir > 1 )
        {
          fprintf( stderr, "FC-Min DF direction must be 0 or 1\n" );
          exit( 1 );
        }
        if ( mo->CD_mutations > 100 )
        {
          fprintf( stderr, "Number of mutations must be less of equal to 100%\n" );
          exit( 1 );
        }
        if ( mo->IR_mutations > 100 )
        {
          fprintf( stderr, "Number of mutations must be less of equal to 100%\n" );
          exit( 1 );
        }
        if ( mo->COV_type > 100 )
        {
          fprintf( stderr, "Unknown CP solution method (%i)\n", mo->COV_type );
          exit( 1 );
        }
        if ( mo->IE_type > 100 )
        {
          fprintf( stderr, "Unknown IE method (%i)\n", mo->IE_type );
          exit( 1 );
        }
        if ( mo->OPT_crit > 5 )
        {
          fprintf( stderr, "Unknown optimization criterion (%i)\n", mo->OPT_crit );
          exit( 1 );
        }

        if ( mo->COV_end == 1 && ( mo->STOP_improvement == 1 || mo->STOP_optimization == 1 ) )
        {
          fprintf( stderr, "Cannot combine endcov parameter with optimization criteria based stopping condition\n" );
          exit( 0 );
        }
        if ( srcf->type != 1 )
        {
          fprintf( stderr, "Function must be fr type\n" );
          exit( 1 );
        }
        if ( mo->STOP_time == 0 && mo->STOP_iterations == 0 && mo->STOP_improvement == 0 && mo->STOP_optimization == 0 )
          mo->STOP_iterations = 1;

        fprintf( stderr, "Minimizing...\n\n" );
        minf = Minimize( srcf );
        if ( minf == 0 )
        {
          fprintf( stderr, "Error occured during minimization\n" );
          return 1;
        }
        minf->GetInfo();
        fprintf( stderr, "Elapsed time: %.3f s\n", info.tm );
        fprintf( stderr, "Iterations: %i\n", info.cyc );
        fprintf( stderr, "Terms: %i\n", minf->info.groupterms );
        fprintf( stderr, "Literals: %i\n", minf->info.liters );
        fprintf( stderr, "Output cost: %i\n", minf->info.ON_terms );
        fprintf( stderr, "Gates: %i\n", minf->info.gates );
        fprintf( stderr, "GEs: %i\n", minf->info.GEs );
        fprintf( stderr, "\n" );
        minf->SaveFnct( onm );
      break;
      case 1:
        srcf->GetInfo();
        printf( "Terms: %i\n", srcf->info.groupterms );
        printf( "Literals: %i\n", srcf->info.liters );
        printf( "Output cost: %i\n", srcf->info.ON_terms );
        printf( "Gates: %i\n", srcf->info.gates );
        printf( "GEs: %i\n", srcf->info.GEs );
        printf( "\n" );
      break;
      case 2:
        if ( CheckConsist( srcf ) != 0 ) fprintf( stderr, "Function is consistent\n" );
        else
          fprintf( stderr, "Inconsistent function!\n" );
      break;
      case 3:
        if ( srcf->type != 1 )
        {
          fprintf( stderr, "Function must be fr type\n" );
          exit( 1 );
        }
        if ( NoIntersections( srcf ) == 0 )
        {
          fprintf( stderr, "Inconsistent function" );
          exit( 1 );
        }
        else
          fprintf( stderr, "OK\n" );
        srcf->SaveFnct( onm );
      break;
      case 4:
        minf = SplitToMinterms( srcf );
        if ( minf == 0 )
        {
          fprintf( stderr, "Inconsistent function" );
          exit( 1 );
        }
        else
          fprintf( stderr, "OK\n" );
        minf->SaveFnct( onm );
      break;
    }
  }
  /* long sec,usec; k_gettime(sec,usec); printf("test : %ld,%ld\n",sec,usec); k_gettime(sec,usec);
  printf("test : %ld,%ld\n",sec,usec); MElement::release(); */
  delete srcf;
  delete minf;

  return 0;
}
