//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef BC
  #include <algorithm.h>
#endif
#include <sys/timeb.h>
#include <time.h>
#include <limits.h>
//---------------------------------------------------------------------------
#include "conf.h"
#include "tpfnct.h"
#include "decomp.h"
#include "dft.h"
#include "bufferstruct.h"
#include "buffers.h"
#include "tboolfnct.h"
#include "algs.h"
#include "kernel.h"
#include "cover.h"
#include "ie.h"
#include "implcheck.h"
#include "cd.h"
#include "fcmin.h"
#include "ir.h"
#include "checks.h"
//---------------------------------------------------------------------------
TBoolFnct::TBoolFnct( int iv, int ov, int vct, int tp )
{
  int i;
  this->terms = vct;
  this->inpvars = iv;
  this->outvars = ov;
  this->type = tp;
  this->fnct = 0;
  this->last = 0;
  this->inames = 0;
  this->onames = 0;
  this->parent = 0;
  this->parent = 0;
  this->split = 0;
  this->inames = new char * [iv];
  for ( i = 0; i < iv; i++ ) this->inames[i] = 0;
  this->onames = new char * [ov];
  for ( i = 0; i < ov; i++ ) this->onames[i] = 0;
  this->decomp = 0;
  this->index = 0;
}
//---------------------------------------------------------------------------
TBoolFnct& TBoolFnct::operator=(TBoolFnct &_fn)
{
  this->terms = _fn.terms;
  this->inpvars = _fn.inpvars;
  this->outvars = _fn.outvars;
  this->type = _fn.type;
  this->info = _fn.info;
  this->firstpass = _fn.firstpass;
  this->hits = _fn.hits;
  this->tries = _fn.tries;
  this->parent = _fn.parent;
  this->last = 0;
  TPfnct *pointer = _fn.fnct;
  while ( pointer != 0 )
  {
    TPfnct *temp = new TPfnct();
    temp->operator=(*pointer);
    if ( this->last != 0 ) this->last->next = temp;
    this->last = temp;
    if ( this->fnct == 0 ) this->fnct = temp;
    this->terms++;
    pointer = pointer->next;
  }
  return *this;
}
//---------------------------------------------------------------------------
TBoolFnct::~TBoolFnct()
{
  TPfnct * fnct2;
  int i;
  if ( this == 0 ) return;
  while ( this->fnct != 0 )
  {
    delete[] this->fnct->iv;
    delete[] this->fnct->ov;
    this->fnct->DeleteCovs();
    fnct2 = this->fnct->next;
    delete this->fnct;
    this->fnct = fnct2;
  }
  if ( this->parent == 0 )
  {
    this->DeleteNames();
    this->CloseSourceFnct();
    //delete []fn->ffbinds;
  }
  if ( this->split == 1 ) this->DeleteSplit();

  if ( this->index != 0 ) delete[] this->index;
  //    delete this;
}
//---------------------------------------------------------------------------
void TBoolFnct::NameIVars()
{
  int i;
  if ( this->inames != 0 )
    for ( i = 0; i < this->inpvars; i++ )
    {
      this->inames[i] = new char[7];
      sprintf( this->inames[i], "x%i", i );
    }
}
//---------------------------------------------------------------------------
void TBoolFnct::NameOVars()
{
  int i;

  if ( this->onames != 0 )
    for ( i = 0; i < this->outvars; i++ )
    {
      this->onames[i] = new char[7];
      sprintf( this->onames[i], "y%i", i );
    }
}
//---------------------------------------------------------------------------
TBoolFnct * TBoolFnct::CreateChildFnct()
{
  TBoolFnct * fnc = new TBoolFnct( this->inpvars, this->outvars, 0, this->type );
  delete[] fnc->inames;
  delete[] fnc->onames;
  fnc->inames = this->inames;
  fnc->onames = this->onames;
  fnc->parent = this;
  fnc->split = 0;
  fnc->decomp = this->decomp;
  return fnc;
}
//---------------------------------------------------------------------------
TPfnct * TBoolFnct::AddEmptyTermToEnd()
{
  TPfnct * t = new TPfnct();
  if ( this->last != 0 ) this->last->next = t;
  this->last = t;
  if ( this->fnct == 0 ) this->fnct = t;
  this->terms++;
  return t;
}
//---------------------------------------------------------------------------
TPfnct * TBoolFnct::AddTermToBeginning( char * ivr, char * ovr, int tdim )
{
  TPfnct * t = new TPfnct(this,ivr,ovr,tdim);
  t->next = this->fnct;
  this->fnct = t;
  this->terms++;
  return t;
}
//---------------------------------------------------------------------------
TPfnct * TBoolFnct::AddTermToEnd( char * ivr, char * ovr, int tdim )
{
  TPfnct * t = new TPfnct(this,ivr,ovr,tdim);
  t->next = 0;
  if ( this->last != 0 ) this->last->next = t;
  this->last = t;
  if ( this->fnct == 0 ) this->fnct = t;
  this->terms++;
  return t;
}
//---------------------------------------------------------------------------
void TBoolFnct::JoinFnct( TBoolFnct * src )
{
  TPfnct *tempFnct = 0;
  TPfnct *prev = 0;
  TPfnct *pointer = src->fnct;
  while ( pointer != 0 )
  {
  TPfnct *temp = new TPfnct();
  temp->operator=(*pointer);
  if ( prev != 0 ) prev->next = temp;
  prev = temp;
  if ( tempFnct == 0 ) tempFnct = temp;
  pointer = pointer->next;
  }

  if ( this->fnct == 0 )
  {
    this->fnct = tempFnct;
    this->last = prev;
  }
  else
  {
    this->last->next = tempFnct;
    if ( src->last != 0 ) this->last = prev;
  }
  this->terms = this->terms + src->terms;
  if ( src->parent == 0 )
  {
    src->DeleteNames();
    src->CloseSourceFnct();
  }
  if ( src->split == 1 ) src->DeleteSplit();
  if ( src->index != 0 ) delete[] src->index;
  delete src;
}
//---------------------------------------------------------------------------
TBoolFnct * TBoolFnct::LoadFnct( char * filn )
{
  FILE * f;
  int i, j;
  char s[65000];
  TBoolFnct * fn = 0;
  int dend;
  char s1[65000], s2[65000];
  int iv = 0, ov = 0, tp = 0;
  char * * inames = 0, * * onames = 0;

  gauge.NewGauge( "Loading PLA file", 0, 1 );
  if ( ( f = fopen( filn, "rt" ) ) == 0 ) return 0;
  dend = 0;
  while ( !feof( f ) && dend == 0 )
  {
    fscanf( f, "%s", s );
    if ( strcmpi( s, ".i" ) == 0 )
    {
      fscanf( f, "%s", s );
      iv = atoi( s );
    }
    else if ( strcmpi( s, ".o" ) == 0 )
    {
      fscanf( f, "%s", s );
      ov = atoi( s );
    }
    else if ( strcmpi( s, ".type" ) == 0 )
    {
      fscanf( f, "%s", s );
      if ( strcmpi( s, "fr" ) == 0 ) tp = 1;
    }
    else if ( strcmpi( s, ".ilb" ) == 0 )
    {
      inames = new char * [iv];
      for ( i = 0; i < iv; i++ )
      {
        fscanf( f, "%s", s );
        inames[i] = new( char[strlen( s ) + 1] );
        strcpy( inames[i], s );
      }
    }
    else if ( strcmpi( s, ".ob" ) == 0 )
    {
      onames = new char * [ov];
      for ( i = 0; i < ov; i++ )
      {
        fscanf( f, "%s", s );
        onames[i] = new( char[strlen( s ) + 1] );
        strcpy( onames[i], s );
      }
    }
    else if ( strncmpi( s, ".e", 2 ) == 0 ) dend = 1;
    else if ( s[0] == '.' )
    {
      do fscanf( f, "%c", s );
      while ( s[0] != 10 && s[0] != 13 );
    }
    else if ( s[0] == '0' || s[0] == '1' || s[0] == '2' || s[0] == '3' || s[0] == '4' || s[0] == '-' || s[0] == '~' )
    {
      if ( fn == 0 )
      {
        fn = new TBoolFnct( iv, ov, 0, tp );
        if ( inames != 0 )
        {
          delete[] fn->inames;
          fn->inames = inames;
        }
        else
          fn->NameIVars();
        if ( onames != 0 )
        {
          delete[] fn->onames;
          fn->onames = onames;
        }
        else
          fn->NameOVars();
      }
      i = 0;
      j = 0;
      while ( i != iv && !feof( f ) )
      {
        if ( ( unsigned )j == strlen( s ) )
        {
          fscanf( f, "%s", s );
          j = 0;
        }
        if ( s[j] == '0' || s[j] == '1' || s[j] == '-' || s[j] == '~' )
          s1[i++] = s[j];
        else if ( s[j] == '2' )
          s1[i++] = '-';
        else if ( s[j] == '3' )
          s1[i++] = '~';
        else if ( s[j] == '4' )
          s1[i++] = '1';
        j++;
      }
      if ( feof( f ) ) break;
      s1[i] = 0;
      i = 0;
      j++;
      while ( i != ov && !feof( f ) )
      {
        if ( ( unsigned )j >= strlen( s ) )
        {
          fscanf( f, "%s", s );
          j = 0;
        }
        if ( s[j] == '0' || s[j] == '1' || s[j] == '-' || s[j] == '~' )
          s2[i++] = s[j];
        else if ( s[j] == '2' )
          s2[i++] = '-';
        else if ( s[j] == '3' )
          s2[i++] = '~';
        else if ( s[j] == '4' )
          s2[i++] = '1';
        j++;
      }
      s2[i] = 0;
      fn->AddTermToEnd( s1, s2, fn->GetDim( s1 ) );
    }
  }
  fclose( f );
  fn->NullFFBinds();
  gauge.CloseGauge();
  return fn;
}
//---------------------------------------------------------------------------
int TBoolFnct::SaveFnct( char * filn )
{
  FILE * f;
  TPfnct * ff;
  int i;

  gauge.NewGauge( "Saving PLA", this->terms, 1 );
  if ( filn != 0 ) f = fopen( filn, "wt" );
  else f = stdout;
  fprintf( f, ".i %i\n", this->inpvars );
  fprintf( f, ".o %i\n", this->outvars );
  fprintf( f, ".p %i\n", this->terms );
  if ( this->inames != 0 && go->NOT_savevarnames )
  {
    fprintf( f, ".ilb" );
    for ( i = 0; i < this->inpvars; i++ ) fprintf( f, " %s", this->inames[i] );
    fprintf( f, "\n" );
  }
  if ( this->onames != 0 && go->NOT_savevarnames )
  {
    fprintf( f, ".ob" );
    for ( i = 0; i < this->outvars; i++ ) fprintf( f, " %s", this->onames[i] );
    fprintf( f, "\n" );
  }
  fprintf( f, ".type " );
  switch ( this->type )
  {
    case 0:
      fprintf( f, "fd\n" );
    break;
    case 1:
      fprintf( f, "fr\n" );
  }
  ff = this->fnct;
  i = 0;
  while ( ff != 0 )
  {
    ff->iv[this->inpvars] = 0;
    ff->ov[this->outvars] = 0;
    fprintf( f, "%s %s\n", ff->iv, ff->ov );
    ff = ff->next;
    if ( gauge.GaugeStep( i++, 0 ) != 0 ) break;
  }
  fprintf( f, ".e\n" );
  fclose( f );
  gauge.CloseGauge();
  return 1;
}
//---------------------------------------------------------------------------
void TBoolFnct::FillEmptyFnct()
{
  int i, j;
  TPfnct * fnct2;
  for ( j = 1; j <= this->terms; j++ )
  {
    fnct2 = new TPfnct;
    fnct2->valid = 1;
    fnct2->prime = 0;
    fnct2->origin = 0;
    fnct2->cov = 0;
    fnct2->covfield = 0;
    fnct2->next = this->fnct;
    if ( this->fnct == 0 ) this->last = fnct2;
    this->fnct = fnct2;
    this->fnct->iv = new( char[this->inpvars + 1] );
    for ( i = 0; i < this->inpvars; i++ ) this->fnct->iv[i] = '-';
    this->fnct->iv[i] = 0;
    this->fnct->ov = new( char[this->outvars + 1] );
    for ( i = 0; i < this->outvars; i++ ) this->fnct->ov[i] = '~';
    this->fnct->ov[i] = 0;
  }
}
//---------------------------------------------------------------------------
void TBoolFnct::GetInfo()
{
  TPfnct * ff;
  int i, j, c;
  int d;
  char * o1;

  o1 = new char[this->outvars];
  setmem( o1, this->outvars, 0 );
  this->info.liters = 0;
  this->info.ON_terms = 0;
  this->info.OFF_terms = 0;
  this->info.DC_terms = 0;
  this->info.groupterms = this->terms;
  this->info.selterms = this->terms;
  this->info.primes = 0;
  this->info.GEs = 0;
  this->info.gates = 0;
  ff = this->fnct;
  while ( ff != 0 )
  {
    d = 0;
    for ( i = 0; i < this->inpvars; i++ )
      if ( ff->iv[i] == '0' || ff->iv[i] == '1' ) d++;
    ff->dim = this->inpvars - d;
    this->info.liters += d;
    this->info.gates += max( d - 1, 0 );
    if ( d > 1 ) this->info.GEs += d;
    for ( i = 0; i < this->outvars; i++ )
      switch ( ff->ov[i] )
      {
        case '0':
          this->info.OFF_terms++;
        break;
        case '1':
          this->info.ON_terms++;
          o1[i] ++;
        break;
        case '-':
          this->info.DC_terms++;
      }
    if ( ff->valid == 0 ) this->info.selterms--;
    if ( ff->prime == 1 ) this->info.primes++;
    ff = ff->next;
  }
  this->info.gates += this->info.ON_terms - this->outvars;
  if ( this->info.gates < 0 ) this->info.gates = 0;
  this->info.GEs += this->info.ON_terms;
  for ( i = 0; i < this->outvars; i++ )
    if ( o1[i] == 1 ) this->info.GEs--;
  delete[] o1;
}
//---------------------------------------------------------------------------
PartInfo TBoolFnct::GetPartinfo()
{
  PartInfo inf;
  int i, j;
  char * inpusg;

  inf.inpsum = 0;
  CountAssignment( this );
  for ( i = 0; i < mo->dopts->blocks; i++ )
    inf.inpsum += this->decomp->icnt[i];
  inpusg = new char[this->inpvars];
  setmem( inpusg, this->inpvars, 0 );
  inf.inpusage = 0;
  inf.branchinps = 0;
  inf.maxload = 0;
  inf.branchings = 0;
  inf.avgload = 0;
  for ( i = 0; i < mo->dopts->blocks; i++ )
    for ( j = 0; j < this->inpvars; j++ )
      if ( this->decomp->inps[i] [j] != 'N' ) inpusg[j] ++;
  for ( j = 0; j < this->inpvars; j++ )
  {
    if ( inpusg[j] != 0 ) inf.inpusage++;
    if ( inpusg[j] > 1 )
    {
      inf.branchinps++;
      inf.branchings += inpusg[j] - 1;
    }
    inf.avgload += inpusg[j];
    if ( inpusg[j] > inf.maxload ) inf.maxload = inpusg[j];
  }
  inf.avgload /= ( double )this->inpvars;
  delete[] inpusg;
  return inf;
}
//---------------------------------------------------------------------------
DFTInfo TBoolFnct::GetDFTinfo()
{
  int i, c;
  DFTInfo inf;

  inf.conesum = 0;
  inf.maxcone = 0;
  for ( i = 0; i < this->outvars; i++ )
  {
    c = ConeSize( this, i );
    inf.conesum += c;
    if ( c > inf.maxcone ) inf.maxcone = c;
  }
  return inf;
}
//---------------------------------------------------------------------------
char * TBoolFnct::FunctToSymbols( int ovr )
{
  TPfnct * ff;
  int i;
  char s[65000];
  int frst, frst2;
  char * s1;
  int tautflg, kontrflg;

  ff = this->fnct;
  strcpy( s, this->onames[ovr] );
  strcat( s, "=" );
  frst = 1;
  kontrflg = 1;
  while ( ff != 0 )
  {
    if ( ff->ov[ovr] == '1' )
    {
      kontrflg = 0;
      if ( frst == 0 ) strcat( s, go->NOT_add );
      frst = 0;
      frst2 = 1;
      tautflg = 1;
      for ( i = 0; i < this->inpvars; i++ )
      {
        if ( ff->iv[i] != '-' )
        {
          tautflg = 0;
          if ( frst2 == 0 ) strcat( s, go->NOT_multiply );
          frst2 = 0;
          if ( ff->iv[i] == '0' && go->NOT_postneg == 0 ) strcat( s, go->NOT_neg );
          strcat( s, this->inames[i] );
          if ( ff->iv[i] == '0' && go->NOT_postneg == 1 ) strcat( s, go->NOT_neg );
        }
      }
      if ( tautflg == 1 ) strcat( s, "1" );
    }
    ff = ff->next;
  }
  if ( kontrflg == 1 ) strcat( s, "0" );
  s1 = new char[strlen( s ) + 1];
  strcpy( s1, s );
  return s1;
}
//---------------------------------------------------------------------------
void TBoolFnct::DeleteInvalidTerms()
{
  TPfnct * ff, * ff1, * ff2;

  ff = this->fnct;
  ff2 = 0;
  while ( ff != 0 )
  {
    if ( ff->valid == 0 )
    {
      if ( this->fnct == ff )
      {
        this->fnct = ff->next;
        if ( this->fnct == 0 ) this->last = 0;
        delete[] ff->iv;
        delete[] ff->ov;
        ff->DeleteCovs();
        delete ff;
        ff = this->fnct;
        ff2 = ff;
        this->terms--;
      }
      else
      {
        ff1 = ff->next;
        delete[] ff->iv;
        delete[] ff->ov;
        ff->DeleteCovs();
        if ( this->last == ff ) this->last = ff2;
        delete ff;
        ff2->next = ff1;
        ff = ff1;
        this->terms--;
      }
    }
    else
    {
      ff2 = ff;
      ff = ff->next;
    }
  }
}
//---------------------------------------------------------------------------
void TBoolFnct::SplitFnct()
{
  TPfnct * ff, * sf;
  int i;

  if ( this->split == 1 ) return;
  this->onterms = new( int[this->outvars] );
  this->offterms = new( int[this->outvars] );
  this->dcterms = new( int[this->outvars] );
  this->onset = new( TPfnct * [this->outvars] );
  this->offset = new( TPfnct * [this->outvars] );
  this->dcset = new( TPfnct * [this->outvars] );
  this->maxonterms = 0;
  for ( i = 0; i < this->outvars; i++ )
  {
    this->onterms[i] = 0;
    this->offterms[i] = 0;
    this->dcterms[i] = 0;
    this->onset[i] = 0;
    this->offset[i] = 0;
    this->dcset[i] = 0;
  }
  ff = this->fnct;
  while ( ff != 0 )
  {
    for ( i = 0; i < this->outvars; i++ )
    {
      if ( ff->ov[i] == '1' )
      {
        sf = new TPfnct;
        sf->cov = 0;
        sf->covfield = 0;
        sf->next = this->onset[i];
        this->onset[i] = sf;
        sf->iv = new char[this->inpvars + 1];
        sf->ov = 0;
        memcpy( sf->iv, ff->iv, this->inpvars );
        sf->iv[this->inpvars] = 0;
        sf->dim = ff->dim;
        sf->cvd = 0;
        this->onterms[i] ++;
        if ( this->maxonterms < this->onterms[i] ) this->maxonterms = this->onterms[i];
      }
      else if ( ff->ov[i] == '0' && this->type == 1 )
      {
        sf = new TPfnct;
        sf->cov = 0;
        sf->covfield = 0;
        sf->next = this->offset[i];
        this->offset[i] = sf;
        sf->iv = new char[this->inpvars + 1];
        sf->ov = 0;
        memcpy( sf->iv, ff->iv, this->inpvars );
        sf->iv[this->inpvars] = 0;
        sf->dim = ff->dim;
        sf->cvd = 0;
        this->offterms[i] ++;
        if ( mo->IC_type == 1 ) buffers->offtree[i]->AddToTreeHistory( ff );
      }
      else if ( ff->ov[i] == '-' && this->type == 0 )
      {
        sf = new TPfnct;
        sf->cov = 0;
        sf->covfield = 0;
        sf->next = this->dcset[i];
        this->dcset[i] = sf;
        sf->iv = new char[this->inpvars + 1];
        sf->ov = 0;
        memcpy( sf->iv, ff->iv, this->inpvars );
        sf->iv[this->inpvars] = 0;
        sf->dim = ff->dim;
        sf->cvd = 0;
        this->dcterms[i] ++;
      }
    }
    ff = ff->next;
  }
  this->split = 1;
}
//---------------------------------------------------------------------------
TPfnct * TBoolFnct::DeleteTerm( TPfnct * ff, TPfnct * pff )
{
  TPfnct * nff;

  nff = ff->next;
  if ( this->fnct == ff ) this->fnct = nff;
  else pff->next = nff;
  ff->DeleteCovs();
  delete[] ff->iv;
  delete[] ff->ov;
  if ( this->last == ff ) this->last = pff;
  delete ff;
  this->terms--;
  return nff;
}
//---------------------------------------------------------------------------
void TBoolFnct::DeleteTerm( int t ) /* Deletes term # t */
{
  TPfnct * ff, * ff1;
  int i;

  if ( this == 0 ) return;
  if ( t == 0 )
  {
    delete[] this->fnct->iv;
    delete[] this->fnct->ov;
    this->fnct->DeleteCovs();
    ff = this->fnct->next;
    delete this->fnct;
    this->fnct = ff;
  }
  else
  {
    ff = this->fnct;
    i = 0;
    while ( ff != 0 && i != t - 1 )
    {
      ff = ff->next;
      i++;
    }
    if ( ff == 0 ) return;
    delete[] ff->next->iv;
    delete[] ff->next->ov;
    ff->next->DeleteCovs();
    ff1 = ff->next->next;
    delete ff->next;
    ff->next = ff1;
    if ( ff1 == 0 ) this->last = ff;
  }
  this->terms--;
}
//---------------------------------------------------------------------------
int TBoolFnct::GetDim( char * ivr )
{
  int i, j;
  j = 0;
  for ( i = 0; i < this->inpvars; i++ )
    if ( ivr[i] == '-' ) j++;
  return j;
}
//---------------------------------------------------------------------------
void TBoolFnct::NullFFBinds()
{
  int i;
  if ( this != 0 )
  {
    this->ffbinds = new int[this->outvars];
    for ( i = 0; i < this->outvars; i++ ) this->ffbinds[i] = -1;
  }
}
//---------------------------------------------------------------------------
void TBoolFnct::SaveFnctAsSymbols( char * filn )
{
  FILE * f;
  int i;

  gauge.NewGauge( "Saving function as text", this->outvars, 1 );
  if ( filn != 0 ) f = fopen( filn, "wt" );
  else f = stdout;
  for ( i = 0; i < this->outvars; i++ )
  {
    fprintf( f, "%s\n", this->FunctToSymbols( i ) );
    if ( gauge.GaugeStep( i, 0 ) != 0 ) break;
  }
  fclose( f );
  gauge.CloseGauge();
}
//---------------------------------------------------------------------------
void TBoolFnct::SaveFnctAsVHDL( char * filn )
{
  FILE * f;
  int i, j, fl;
  char n[255];
  char * np;

  gauge.NewGauge( "Saving function as VHDL", 0, 1 );
  this->Reindex();
  if ( filn != 0 ) f = fopen( filn, "wt" );
  else f = stdout;
  strcpy( n, filn );
  np = strrchr( n, '\\' ) + 1;
  strchr( np, '.' ) [0] = 0;
  fprintf( f, "library IEEE;\n" );
  fprintf( f, "use IEEE.std_logic_1164.all;\n\n" );
  fprintf( f, "entity %s is\n", np );
  fprintf( f, "   port(" );
  for ( i = 0; i < this->inpvars; i++ )
    if ( i == 0 ) fprintf( f, "%s", this->inames[i] );
    else
      fprintf( f, ", %s", this->inames[i] );
  fprintf( f, " : in std_logic;\n" );
  fprintf( f, "      " );
  for ( i = 0; i < this->outvars; i++ )
    if ( i == 0 ) fprintf( f, "%s", this->onames[i] );
    else
      fprintf( f, ", %s", this->onames[i] );
  fprintf( f, " : out std_logic);\n\n" );
  fprintf( f, "begin\n" );
  fprintf( f, "end %s;\n", np );
  fprintf( f, "\n" );
  fprintf( f, "architecture %s_arch of %s is\n", np, np );
  fprintf( f, "   signal " );
  for ( i = 0; i < this->terms; i++ )
    if ( i == 0 ) fprintf( f, "t%i", i );
    else
      fprintf( f, ", t%i", i );
  fprintf( f, " : std_logic;\n" );
  fprintf( f, "   signal " );
  for ( i = 0; i < this->inpvars; i++ )
    if ( i == 0 ) fprintf( f, "n_%s", this->inames[i] );
    else
      fprintf( f, ", n_%s", this->inames[i] );
  fprintf( f, " : std_logic;\n" );
  fprintf( f, "\n" );
  fprintf( f, "begin\n" );
  for ( i = 0; i < this->inpvars; i++ )
    fprintf( f, "   n_%s <= not %s;\n", this->inames[i], this->inames[i] );
  fprintf( f, "\n" );
  for ( i = 0; i < this->terms; i++ )
  {
    fprintf( f, "   t%i <= ", i );
    fl = 0;
    for ( j = 0; j < this->inpvars; j++ )
      if ( this->index[i]->iv[j] != '-' )
      {
        if ( fl != 0 ) fprintf( f, " and " );
        if ( this->index[i]->iv[j] == '0' ) fprintf( f, "n_" );
        fl = 1;
        fprintf( f, "%s", this->inames[j] );
      }
    fprintf( f, ";\n" );
  }
  fprintf( f, "\n" );
  for ( i = 0; i < this->outvars; i++ )
  {
    fprintf( f, "   %s <= ", this->onames[i] );
    fl = 0;
    for ( j = 0; j < this->terms; j++ )
      if ( this->index[j]->ov[i] == '1' )
      {
        if ( fl != 0 ) fprintf( f, " or " );
        fprintf( f, "t%i", j );
        fl = 1;
      }
    if ( fl == 0 ) fprintf( f, "0" );
    fprintf( f, ";\n" );
  }
  fprintf( f, "end %s_arch\n", np );

  fclose( f );
  gauge.CloseGauge();
}
//---------------------------------------------------------------------------
void TBoolFnct::SaveFnctAsVerilog( char * filn )
{
  FILE * f;
  int i, j, fl;
  char n[255];
  char * np;

  gauge.NewGauge( "Saving function as Verilog", 0, 1 );
  this->Reindex();
  if ( filn != 0 ) f = fopen( filn, "wt" );
  else f = stdout;
  strcpy( n, filn );
  np = strrchr( n, '\\' ) + 1;
  strchr( np, '.' ) [0] = 0;

  fprintf( f, "module %s (", np );
  for ( i = 0; i < this->inpvars; i++ )
  {
    if ( i != 0 ) fprintf( f, ", " );
    fprintf( f, "%s", this->inames[i] );
  }
  for ( i = 0; i < this->outvars; i++ ) fprintf( f, ", %s", this->onames[i] );
  fprintf( f, ");\n" );
  fprintf( f, "input " );
  for ( i = 0; i < this->inpvars; i++ )
  {
    if ( i != 0 ) fprintf( f, ", " );
    fprintf( f, "%s", this->inames[i] );
  }
  fprintf( f, ";\n" );
  fprintf( f, "output " );
  for ( i = 0; i < this->outvars; i++ )
  {
    if ( i != 0 ) fprintf( f, ", " );
    fprintf( f, "%s", this->onames[i] );
  }
  fprintf( f, ";\n" );
  fprintf( f, "wire " );
  for ( i = 0; i < this->terms; i++ )
  {
    if ( i != 0 ) fprintf( f, ", " );
    fprintf( f, "t%i", i );
  }
  fprintf( f, ";\n\n" );

  fprintf( f, "assign\n" );
  for ( i = 0; i < this->terms; i++ )
  {
    fprintf( f, "   t%i = ", i );
    fl = 0;
    for ( j = 0; j < this->inpvars; j++ )
      if ( this->index[i]->iv[j] != '-' )
      {
        if ( fl != 0 ) fprintf( f, " & " );
        fl = 1;
        if ( this->index[i]->iv[j] == '0' ) fprintf( f, "~" );
        fprintf( f, "%s", this->inames[j] );
      }
    if ( i != this->terms ) fprintf( f, ",\n" );
    else fprintf( f, ";\n" );
  }
  fprintf( f, "\n" );
  for ( i = 0; i < this->outvars; i++ )
  {
    fprintf( f, "   %s = ", this->onames[i] );
    fl = 0;
    for ( j = 0; j < this->terms; j++ )
      if ( this->index[j]->ov[i] == '1' )
      {
        if ( fl != 0 ) fprintf( f, " | " );
        fl = 1;
        fprintf( f, "t%i", j );
      }
    if ( fl == 0 ) fprintf( f, "0" );
    if ( i != this->outvars - 1 ) fprintf( f, ",\n" );
    else fprintf( f, ";\n" );
  }
  fprintf( f, "endmodule\n" );


  fclose( f );
  gauge.CloseGauge();
}
//---------------------------------------------------------------------------
void TBoolFnct::SaveFnctAsVerilog2( char * filn ) /* Doesn't take into account group terms - just equations */
{
  FILE * f;
  int i, j, k, fl, fl2;
  char n[255];
  char * np;

  gauge.NewGauge( "Saving function as Verilog", 0, 1 );
  this->Reindex();
  if ( filn != 0 ) f = fopen( filn, "wt" );
  else f = stdout;
  strcpy( n, filn );
  np = strrchr( n, '\\' ) + 1;
  strchr( np, '.' ) [0] = 0;

  fprintf( f, "module %s (", np );
  for ( i = 0; i < this->inpvars; i++ )
  {
    if ( i != 0 ) fprintf( f, ", " );
    fprintf( f, "%s", this->inames[i] );
  }
  for ( i = 0; i < this->outvars; i++ ) fprintf( f, ", %s", this->onames[i] );
  fprintf( f, ");\n" );
  fprintf( f, "input " );
  for ( i = 0; i < this->inpvars; i++ )
  {
    if ( i != 0 ) fprintf( f, ", " );
    fprintf( f, "%s", this->inames[i] );
  }
  fprintf( f, ";\n" );
  fprintf( f, "output " );
  for ( i = 0; i < this->outvars; i++ )
  {
    if ( i != 0 ) fprintf( f, ", " );
    fprintf( f, "%s", this->onames[i] );
  }
  fprintf( f, ";\n\n" );
  fprintf( f, "assign\n" );

  for ( i = 0; i < this->outvars; i++ )
  {
    fprintf( f, "   %s = ", this->onames[i] );
    fl = 0;
    for ( j = 0; j < this->terms; j++ )
      if ( this->index[j]->ov[i] == '1' )
      {
        if ( fl != 0 ) fprintf( f, " | " );
        fl = 1;
        fprintf( f, "(" );
        fl2 = 0;
        for ( k = 0; k < this->inpvars; k++ )
          if ( this->index[j]->iv[k] != '-' )
          {
            if ( fl2 != 0 ) fprintf( f, " & " );
            fl2 = 1;
            if ( this->index[j]->iv[k] == '0' ) fprintf( f, "~" );
            fprintf( f, "%s", this->inames[k] );
          }
        fprintf( f, ")" );
      }
    if ( fl == 0 ) fprintf( f, "0" );
    if ( i != this->outvars - 1 ) fprintf( f, ",\n" );
    else fprintf( f, ";\n" );
  }
  fprintf( f, "endmodule\n" );
  fclose( f );
  gauge.CloseGauge();
}
//---------------------------------------------------------------------------
void TBoolFnct::SaveFnctAsHTML( char * filn )
{
  FILE * f;
  int i;
  TPfnct * ff;

  gauge.NewGauge( "Saving functin as HTML", this->outvars, 1 );
  if ( filn != 0 ) f = fopen( filn, "wt" );
  else f = stdout;
  fprintf( f, "<html>\n" );
  fprintf( f, "<head>\n" );
  fprintf( f, "<meta http-equiv=""Content-Type"" content=""text/html; charset=iso-8859-1"">\n" );
  fprintf( f, "<body>\n" );
  fprintf( f, "<table border=1>\n" );
  fprintf( f, "<tr>\n" );
  for ( i = 0; i < this->inpvars; i++ )
    fprintf( f, "  <td>%s</td>\n", this->inames[i] );
  for ( i = 0; i < this->outvars; i++ )
    fprintf( f, "  <td bgcolor=""cyan"">%s</td>\n", this->onames[i] );
  fprintf( f, "</tr>\n" );
  ff = this->fnct;
  i = 0;
  while ( ff != 0 )
  {
    fprintf( f, "<tr>\n" );
    for ( i = 0; i < this->inpvars; i++ )
      fprintf( f, "  <td>%c</td>\n", ff->iv[i] );
    for ( i = 0; i < this->outvars; i++ )
      fprintf( f, "  <td bgcolor=""cyan"">%c</td>\n", ff->ov[i] );
    fprintf( f, "</tr>\n" );
    if ( gauge.GaugeStep( i++, 0 ) != 0 ) break;
    ff = ff->next;
  }
  fprintf( f, "</body>\n" );
  fprintf( f, "</html>\n" );
  fclose( f );
  gauge.CloseGauge();
}
//---------------------------------------------------------------------------
void TBoolFnct::DuplNames()
{
  int i;
  char * * n;

  n = this->parent->inames;
  this->inames = new char * [this->inpvars];
  for ( i = 0; i < this->inpvars; i++ )
  {
    this->inames[i] = new char[strlen( n[i] ) + 1];
    strcpy( this->inames[i], n[i] );
  }
  n = this->parent->onames;
  this->onames = new char * [this->outvars];
  for ( i = 0; i < this->outvars; i++ )
  {
    this->onames[i] = new char[strlen( n[i] ) + 1];
    strcpy( this->onames[i], n[i] );
  }
}
//---------------------------------------------------------------------------
void TBoolFnct::DeleteInpVariable( int v )
{
  TPfnct * ff;
  int i;

  if ( this == 0 ) return;
  if ( v >= this->inpvars ) return;
  ff = this->fnct;
  while ( ff != 0 )
  {
    memcpy( ff->iv + v, ff->iv + v + 1, this->inpvars - v );
    ff->iv[this->inpvars - 1] = 0;
    ff = ff->next;
  }
  delete[] this->inames[v];
  for ( i = v; i < this->inpvars - 1; i++ )
    this->inames[i] = this->inames[i + 1];
  this->inpvars--;
}
//---------------------------------------------------------------------------
void TBoolFnct::DeleteOutVariable( int v )
{
  TPfnct * ff;
  int i;

  if ( this == 0 ) return;
  if ( v >= this->outvars ) return;
  ff = this->fnct;
  while ( ff != 0 )
  {
    memcpy( ff->ov + v, ff->ov + v + 1, this->outvars - v );
    ff->ov[this->outvars - 1] = 0;
    ff = ff->next;
  }
  delete[] this->onames[v];
  for ( i = v; i < this->outvars - 1; i++ )
    this->onames[i] = this->onames[i + 1];
  this->outvars--;
}
//---------------------------------------------------------------------------
void TBoolFnct::InsertTerm( int t )
{
  TPfnct * ff, * ff1;
  int i;

  if ( this == 0 ) return;

  ff = new TPfnct;
  ff->valid = 1;
  ff->cov = 0;
  ff->covfield = 0;
  ff->dim = 0;
  ff->prime = 0;
  ff->origin = 0;
  ff->iv = new char[this->inpvars + 1];
  setmem( ff->iv, this->inpvars, '-' );
  ff->iv[this->inpvars] = 0;
  ff->ov = new char[this->outvars + 1];
  setmem( ff->ov, this->outvars, '~' );
  ff->ov[this->outvars] = 0;
  if ( t == 0 )
  {
    ff->next = this->fnct;
    this->fnct = ff;
  }
  else
  {
    ff1 = this->fnct;
    i = 0;
    while ( ff1 != 0 && i != t - 1 )
    {
      ff1 = ff1->next;
      i++;
    }
    ff->next = ff1->next;
    ff1->next = ff;
  }
  this->terms++;
}
//---------------------------------------------------------------------------
void TBoolFnct::InsertInpVariable( int v )
{
  TPfnct * ff;
  int i;
  char * iv;
  char * * in;

  if ( this == 0 ) return;
  if ( v >= this->inpvars ) return;
  ff = this->fnct;
  while ( ff != 0 )
  {
    iv = new char[this->inpvars + 2];
    memcpy( iv, ff->iv, v );
    iv[v] = '-';
    memcpy( iv + v + 1, ff->iv + v, this->inpvars - v );
    iv[this->inpvars + 1] = 0;
    delete[] ff->iv;
    ff->iv = iv;
    ff = ff->next;
  }
  in = new char * [this->inpvars + 2];
  for ( i = 0; i < v; i++ )
    in[i] = this->inames[i];
  for ( i = v; i < this->inpvars; i++ )
    in[i + 1] = this->inames[i];
  in[v] = new char[255];
  strcpy( in[v], "new" );
  delete[] this->inames;
  this->inames = in;
  this->inpvars++;
}
//---------------------------------------------------------------------------
void TBoolFnct::InsertOutVariable( int v )
{
  TPfnct * ff;
  int i;
  char * ov;
  char * * on;

  if ( this == 0 ) return;
  if ( v >= this->outvars ) return;
  ff = this->fnct;
  while ( ff != 0 )
  {
    ov = new char[this->outvars + 2];
    memcpy( ov, ff->ov, v );
    ov[v] = '~';
    memcpy( ov + v + 1, ff->ov + v, this->outvars - v );
    ov[this->outvars + 1] = 0;
    delete[] ff->ov;
    ff->ov = ov;
    ff = ff->next;
  }
  on = new char * [this->outvars + 2];
  for ( i = 0; i < v; i++ )
    on[i] = this->onames[i];
  for ( i = v; i < this->outvars; i++ )
    on[i + 1] = this->onames[i];
  on[v] = new char[255];
  strcpy( on[v], "new" );
  delete[] this->onames;
  this->onames = on;
  this->outvars++;
}
//---------------------------------------------------------------------------
int TBoolFnct::CountFnctHamming( char * c )
{
  int d, h;
  TPfnct * ff;

  d = INT_MAX;
  ff = this->fnct;
  while ( ff != 0 && d > 0 )
  {
    h = CountHamming( ff->iv, c, this->inpvars );
    d = min( d, h );
    ff = ff->next;
  }
  return d;
}
//---------------------------------------------------------------------------
void TBoolFnct::GetOptVal() // extracts optimization value
{
  switch ( mo->OPT_crit )
  {
    case 0:
      this->optval = this->info.groupterms;
    break;
    case 1:
      this->optval = this->info.liters;
    break;
    case 2:
      this->optval = this->info.ON_terms;
    break;
    case 3:
      this->optval = this->info.liters + this->info.ON_terms;
    break;
    case 4:
      this->optval = this->info.gates;
    break;
    case 5:
      this->optval = this->info.GEs;
    break;
      /* case 6: info.inf.inpsum; break; case 7: inf.conesum; break; */
  }
}
//---------------------------------------------------------------------------
void TBoolFnct::Reindex()
{
  TPfnct * ff;
  int i;

  delete[] this->index;
  this->index = new TPfnct * [this->terms];
  ff = this->fnct;
  i = 0;
  while ( ff != 0 )
  {
    this->index[i] = ff;
    ff->index = i;
    ff = ff->next;
    i++;
  }
}
//---------------------------------------------------------------------------
TBoolFnct * TBoolFnct::DuplFnct() /* Creates a duplicate of function. */
{
  TBoolFnct * ofn;
  TPfnct * ff, * ff2;

  ofn = new TBoolFnct( this->inpvars, this->outvars, 0, this->type );
  ff = this->fnct;
  while ( ff != 0 )
  {
    ff2 = ofn->AddTermToEnd( ff->iv, ff->ov, ff->dim );
    ff2->cov = DuplCov( ff->cov );
    ff2->covlen = ff->covlen;
    if ( ff->covfield != 0 )
    {
      ff2->covfield = new int[ff->covlen];
      memcpy( ff2->covfield, ff->covfield, ff->covlen * sizeof( int ) );
    }
    ff = ff->next;
  }
  ofn->parent = this->parent;
  return ofn;
}
//---------------------------------------------------------------------------
void TBoolFnct::SubstOutDCs() /* Substitues output DCs with "0"s */
{
  int i;
  TPfnct * ff;

  ff = this->fnct;
  while ( ff != 0 )
  {
    for ( i = 0; i < this->outvars; i++ )
      if ( ff->ov[i] == '-' ) ff->ov[i] = '0';
    ff = ff->next;
  }
}
//---------------------------------------------------------------------------
void TBoolFnct::MakeCovField( TPfnct * ff )
{
  TCovs * c;
  int d;

  if ( ff->covfield == 0 )
  {
    ff->covfield = new int[ff->covlen];
    c = ff->cov;
    d = 0;
    while ( c != 0 )
    {
      ff->covfield[d++] = c->index + this->parent->maxonterms * c->var;
      c = c->next;
    }
    QuickSort( ff->covfield, d );
  }
}
//---------------------------------------------------------------------------
char * TBoolFnct::IsPrime( TPfnct * ff ) /* Checks if ff is a PI of any output function. srcf = parent function
returns 0 of not, or a pointer to field[ovr] filed[i] = '1' if it is a PI of i output part of ff must be valid */
{
  int i, j;
  char * ispi;
  int ipi;
  char tmp;
  int t;

  ispi = new char[this->outvars];
  setmem( ispi, this->outvars, '1' );
  ipi = 0;

  for ( i = 0; i < this->outvars; i++ )
    if ( ff->ov[i] == '1' )
    {
      for ( j = 0; j < this->inpvars; j++ )
        if ( ff->iv[j] != '-' )
        {
          tmp = ff->iv[j];
          ff->iv[j] = '-';
          t = ImplCheck( this, ff, i );
          ff->iv[j] = tmp;
          if ( t )
          {
            ispi[j] = '0';
            break;
          }
        }
      if ( j == this->inpvars ) ipi = 1;
    }
  if ( !ipi )
  {
    delete[] ispi;
    ispi = 0;
  }
  return ispi;
}
//---------------------------------------------------------------------------
int TBoolFnct::AdvCountPrimes( int mask ) /* counts PI in fn, for a given origin mask */
{
  int i;
  TPfnct * ff;
  char * p;

  i = 0;
  ff = this->fnct;
  while ( ff != 0 )
  {
    if ( ( ff->origin & mask ) == mask )
    {
      p = this->parent->IsPrime( ff );
      if ( p != 0 )
      {
        delete[] p;
        i++;
      }
    }
    ff = ff->next;
  }
  return i;
}
//---------------------------------------------------------------------------
void TBoolFnct::DeleteNames() /* Deletes names of the function */
{
  int i;
  if ( this->inames != 0 )
  {
    for ( i = 0; i < this->inpvars; i++ ) delete[] this->inames[i];
    delete[] this->inames;
  }
  if ( this->onames != 0 )
  {
    for ( i = 0; i < this->outvars; i++ ) delete[] this->onames[i];
    delete[] this->onames;
  }
  this->inames = 0;
  this->onames = 0;
}
//---------------------------------------------------------------------------
void TBoolFnct::DeleteSplit() /* Destroys function split info. */
{
  int i;
  TPfnct * fnct2;

  if ( !this->split ) return;

  for ( i = 0; i < this->outvars; i++ )
  {
    while ( this->onset[i] != 0 )
    {
      delete[] this->onset[i]->iv;
      fnct2 = this->onset[i]->next;
      delete this->onset[i];
      this->onset[i] = fnct2;
    }
    while ( this->offset[i] != 0 )
    {
      delete[] this->offset[i]->iv;
      fnct2 = this->offset[i]->next;
      delete this->offset[i];
      this->offset[i] = fnct2;
    }
    while ( this->dcset[i] != 0 )
    {
      delete[] this->dcset[i]->iv;
      fnct2 = this->dcset[i]->next;
      delete this->dcset[i];
      this->dcset[i] = fnct2;
    }
  }
  delete[] this->onset;
  delete[] this->offset;
  delete[] this->dcset;
  delete[] this->onterms;
  delete[] this->offterms;
  delete[] this->dcterms;
  this->split = 0;
}
//---------------------------------------------------------------------------
void TBoolFnct::InitSourceFnct()
{
  InitDecompAss( this );
}
//---------------------------------------------------------------------------
void TBoolFnct::CloseSourceFnct()
{
  DestroyDecompAss( this );
}
//---------------------------------------------------------------------------
TBoolFnct * TBoolFnct::DumpOnset()
{
  TBoolFnct * ofn;
  TPfnct * aktterm;
  TPfnct * sf;
  int j;
  int ovr;

  ofn = this->CreateChildFnct();
  ofn->type = 0;

  for ( ovr = 0; ovr < this->outvars; ovr++ )
  {
    sf = this->onset[ovr];
    while ( sf != 0 )
    {
      aktterm = ofn->AddEmptyTermToEnd();
      aktterm->iv = new char[ofn->inpvars + 1];
      memcpy( aktterm->iv, sf->iv, this->inpvars );
      aktterm->iv[this->inpvars] = 0;
      aktterm->dim = sf->dim;
      aktterm->ov = new char[ofn->outvars + 1];
      aktterm->ov[ofn->outvars] = 0;
      aktterm->dvar = ovr;
      AddToHistory( buffers->e_buffer, aktterm );
      for ( j = 0; j < this->inpvars; j++ ) buffers->e_buffer->last->ov[j] = '0';
      buffers->e_buffer->last->ov[j] = 0;
      sf = sf->next;
    }
  }
  return ofn;
}
//------------------------------------------------------------------------------
TBoolFnct * OneMinStep( TBoolFnct * srcf, TBoolFnct * allf )
{
  fprintf ( stderr, "CDmin step started .. \n\n" );

  int i;
  TBoolFnct * minf1;

  gauge.NewGauge( "Processing output variables", srcf->outvars, 1 );

  for ( i = 0; i < srcf->outvars; i++ )
  {
#if (mintrck > 0)
    ftime( & tr_t );
#endif
    if ( mo->CD_no == 0 )
    {
      minf1 = CD_Min( srcf, i );
      delete minf1;
    }
    else if ( mo->IE_type == 0 || srcf->firstpass == 1 )
    {
      srcf->DumpOnset();
      srcf->firstpass = 0;
    }
#if (mintrck > 0)
    ftime( & tr_t1 );
    cd_time += ( ( double )tr_t1.time + ( double )tr_t1.millitm / 1000 )
         - ( ( double )tr_t.time + ( double )tr_t.millitm / 1000 );
    tr_t = tr_t1;
#endif
    minf1 = ImplExpand( buffers->e_buffer, i );
#if (mintrck > 0)
    ftime( & tr_t1 );
    exp_time += ( ( double )tr_t1.time + ( double )tr_t1.millitm / 1000 )
         - ( ( double )tr_t.time + ( double )tr_t.millitm / 1000 );
    tr_t = tr_t1;
#endif
    ConstructValidOutputs( minf1 );
#if (mintrck > 0)
    ftime( & tr_t );
#endif
    if ( minf1->outvars > 1 && mo->IR_no == 0 && mo->COV_somin == 0 ) MoReduce( minf1 );
#if (mintrck > 0)
    ftime( & tr_t1 );
    red_time += ( ( double )tr_t1.time + ( double )tr_t1.millitm / 1000 )
         - ( ( double )tr_t.time + ( double )tr_t.millitm / 1000 );
    tr_t = tr_t1;
#endif
    FindCover( minf1 );
    if ( mo->DOM_type > 0 ) SelfDominChk( minf1 );
    //SaveFnct(minf1,"prim.pla");
    if ( allf == 0 )
    {
      allf = minf1;
      //allf = new TBoolFnct( minf1->inpvars, minf1->outvars, 0, minf1->type );
      //allf->operator=(*minf1);
    }
    else
    {
      if ( mo->DOM_type > 1 )
      {
        DominChk( allf, minf1 );
        DominChk( minf1, allf );
      }
      allf->JoinFnct( minf1 );
      if ( mo->DOM_type == 1 ) DuplChk( allf );
    }
#if (mintrck > 0)
    ftime( & tr_t1 );
    dom_time += ( ( double )tr_t1.time + ( double )tr_t1.millitm / 1000 )
         - ( ( double )tr_t.time + ( double )tr_t.millitm / 1000 );
    tr_t = tr_t1;
#endif
    gauge.GaugeStep( i, 0 );
  }
  gauge.CloseGauge();
  return allf;
}
//---------------------------------------------------------------------------
TBoolFnct * OneFCMinStep( TBoolFnct * srcf, TBoolFnct * allf )
{
  fprintf ( stderr, "FCmin step started .. \n\n" );

  TBoolFnct * minf1;

  minf1 = FCMinimize( srcf );
  //  SubstOutDCs(minf1);
  FindCover( minf1 );
  if ( allf == 0 )
  {
    allf = minf1;
  }
  else
  {
    if ( mo->DOM_type > 1 )
    {
      DominChk( allf, minf1 );
      DominChk( minf1, allf );
    }
    allf->JoinFnct( minf1 );
    if ( mo->DOM_type == 1 ) DuplChk( allf );
  }
  return allf;
}

TBoolFnct * Minimize( TBoolFnct * srcf )
{
  struct timeb t, t1;
  FnctInfo inf1;

  TBoolFnct * minf1, * minf;
  time_t tr;
  int stop;
  int crit;
  int lpass, lpass2;
  int dtrm;
  int i;
  int c1, c2, c3, st_tmp;

  buffers = new TBuffers(srcf);

#if (mintrck > 0)
  int ch[30];
  FILE * f, * f1, * f2, * f3;
#endif

  // decomp
  mo->DECOMP_enable = 0;
  mo->DECOMP_CDenable = 0;
  mo->DECOMP_COVenable = 0;
  InitDecompAss( srcf );

  srand(static_cast<unsigned>(time(0)));
  //srand( ( unsigned )time( & tr ) );
  //srand(2);
#if (mintrck > 0)
  f = fopen( "mintrck.dat", "wt" );
  f1 = fopen( "mintrck1.dat", "wt" );
  f2 = fopen( "mintrck_comp.dat", "wt" );
  f3 = fopen( "mintrck_distr.dat", "wt" );
  fprintf( f, "Cycle CD_time IE_time IR_time DC_time CP_time Sp_time Time Lits OC Terms All_terms Sol_PIs\n" );
  fprintf( f1, "Cycle CD_time +IE_time +IR_time +DC_time +CP_time +Sp_time Time Lits OC Terms All_terms Sol_PIs\n" );
  fprintf( f2, "Cycle All_terms All_PIs All_FCs All_comms CurSol_terms CurSol_PIs CurSol_FCs CurSol_comms BestSol_terms BestSol_PIs BestSol_FCs BestSol_comms CurSol_lits CurSol_OC CurSol_GEs BestSol_lits BestSol_OC BestSol_GEs\n" );
  fprintf( f3, "Cycle [All terms] [Current Solution] [Best Solution]\n" );
  fprintf( f3, "BOOM_PIs common_PIs FC_PIs BOOM_npi common_npi FC_npi\n" );
  fprintf( f3, "allf_BOOM_PIs allf_common_PIs allf_FC_PIs allf_BOOM_npi allf_common_npi allf_FC_npi cursol_BOOM_PIs cursol_common_PIs cursol_FC_PIs cursol_BOOM_npi cursol_common_npi cursol_FC_npi bestsol_BOOM_PIs bestsol_common_PIs bestsol_FC_PIs bestsol_BOOM_npi bestsol_common_npi bestsol_FC_npi\n" );
#endif

  info.cyc = 0;
  //    info.allf = 0;
  //    info.minf = 0;
  minf = 0;


  srcf->SplitFnct();
  //MinimizeOffset(ssrcf);

  ftime( & t );
  lpass = 1;
  lpass2 = 1;
  srcf->firstpass = 1;
#if (mintrck > 0)
  cd_time = 0;
  exp_time = 0;
  red_time = 0;
  dom_time = 0;
  cov_time = 0;
  sprs_time = 0;
  ftime( & tr_t );
#endif
  if ( mo->DECOMP_enable == 1 )
    if ( AssignValidOutputs( srcf ) == 1 ) return 0;

  if ( mo->STOP_time == 1 && mo->STOP_iterations == 0 )
    gauge.NewGauge( "Minimizing using BOOM...", mo->STOP_maxtime, 1 );
  else if ( mo->STOP_iterations == 1 && mo->STOP_time == 0 )
    gauge.NewGauge( "Minimizing...", mo->STOP_maxiterations, 1 );
  else
    gauge.NewGauge( "Minimizing...", 0, 1 );
  do
  {
    if ( mo->DECOMP_enable == 1 )
    {
      RandomAssignValidOutputs( srcf );
      CleanInpAssignment( srcf );
    }
    info.cyc++;
    if ( buffers->r_buffer == 0 ) dtrm = 0;
    else dtrm = buffers->r_buffer->terms;
    if ( ( random( 100 ) + 1 ) > mo->CD_FC ) buffers->r_buffer = OneMinStep( srcf, buffers->r_buffer );
    else
      buffers->r_buffer = OneFCMinStep( srcf, buffers->r_buffer );
    dtrm = buffers->r_buffer->terms - dtrm;
#if (mintrck > 0)
    ftime( & tr_t );
#endif
    if ( mo->COV_end == 0 && dtrm > 0 )
    {
      if ( mo->DECOMP_enable == 1 ) minf1 = BlockCover( buffers->r_buffer );
      else
      {
        if ( mo->COV_somin == 0 ) minf1 = Cover( buffers->r_buffer );
        else
          minf1 = SOCover( buffers->r_buffer );
      }
    }
    info.tries = buffers->r_buffer->tries;
    info.F_hits = buffers->r_buffer->hits;
#if (mintrck > 0)
    ftime( & tr_t1 );
    cov_time += ( ( double )tr_t1.time + ( double )tr_t1.millitm / 1000 )
         - ( ( double )tr_t.time + ( double )tr_t.millitm / 1000 );
    tr_t = tr_t1;
#endif
    if ( mo->OR_no == 0 && srcf->outvars > 1 && mo->COV_end == 0 && mo->COV_somin == 0 && dtrm > 0 ) SparseOutput( minf1 );
#if (mintrck > 0)
    ftime( & tr_t1 );
    sprs_time += ( ( double )tr_t1.time + ( double )tr_t1.millitm / 1000 )
         - ( ( double )tr_t.time + ( double )tr_t.millitm / 1000 );
    tr_t = tr_t1;
#endif
    //EssentialCheck(minf1,i,mo);
#if (mintrck > 0)
    if ( minf == 0 )
    {
      st_tmp = 1;
      minf = minf1;
    }
    else
      st_tmp = 0;
    if ( dtrm > 0 && info.cyc > 1 )
    {
      GetInfo( minf1 );
      GetInfo( minf );
      fprintf( f2, "%i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i\n", info.cyc, info.allf->terms, CountPrimes( info.allf->fnct ),
           CountFCs( info.allf->fnct ), CountCommons( info.allf->fnct ), minf1->terms, CountPrimes( minf1->fnct ), CountFCs( minf1->fnct ),
           CountCommons( minf1->fnct ), minf->terms, CountPrimes( minf->fnct ), CountFCs( minf->fnct ), CountCommons( minf->fnct ),
           minf1->info.liters, minf1->info.ON_terms, minf1->info.GEs, minf->info.liters, minf->info.ON_terms, minf->info.GEs );
      fflush( f2 );
      ch[1] = AdvCountPrimes( info.allf, 3 );
      ch[0] = AdvCountPrimes( buffers->r_buffer, 1 ) - ch[1];
      ch[2] = AdvCountPrimes( buffers->r_buffer, 2 ) - ch[1];
      ch[4] = CountCommons( buffers->r_buffer->fnct ) - ch[1];
      ch[5] = CountFCs( buffers->r_buffer->fnct ) - ch[1] - ch[2] - ch[4];
      ch[3] = buffers->r_buffer->terms - ch[0] - ch[1] - ch[2] - ch[4] - ch[5];
      ch[11] = AdvCountPrimes( minf1, 3 );
      ch[10] = AdvCountPrimes( minf1, 1 ) - ch[11];
      ch[12] = AdvCountPrimes( minf1, 2 ) - ch[11];
      ch[14] = CountCommons( minf1->fnct ) - ch[11];
      ch[15] = CountFCs( minf1->fnct ) - ch[11] - ch[12] - ch[14];
      ch[13] = minf1->terms - ch[10] - ch[11] - ch[12] - ch[14] - ch[15];
      ch[21] = AdvCountPrimes( minf, 3 );
      ch[20] = AdvCountPrimes( minf, 1 ) - ch[21];
      ch[22] = AdvCountPrimes( minf, 2 ) - ch[21];
      ch[24] = CountCommons( minf->fnct ) - ch[21];
      ch[25] = CountFCs( minf->fnct ) - ch[21] - ch[22] - ch[24];
      ch[23] = minf->terms - ch[20] - ch[21] - ch[22] - ch[24] - ch[25];
      fprintf( f3, "%i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i\n", info.cyc, ch[0], ch[1], ch[2], ch[3], ch[4],
           ch[5], ch[10], ch[11], ch[12], ch[13], ch[14], ch[15], ch[20], ch[21], ch[22], ch[23], ch[24], ch[25] );
      fflush( f3 );
    }
    if ( st_tmp ) minf = 0;
#endif
    if ( mo->COV_end != 1 && dtrm > 0 )
    {
      minf1->GetInfo();
      minf1->GetOptVal();
      if ( minf == 0 )
      {
        minf = minf1;
        buffers->minf = minf;
        if ( mo->DECOMP_enable == 1 ) minf->decomp = DuplicateDecompAss( srcf );
      }
      else
      {
        if ( minf1->optval < minf->optval )
        {
          if ( mo->DECOMP_enable == 1 ) DestroyDecompAss( minf );
          delete minf;
          minf = minf1;
          buffers->minf = minf;
          if ( mo->DECOMP_enable == 1 ) minf->decomp = DuplicateDecompAss( srcf );
          lpass2 = info.cyc - lpass;
          lpass = info.cyc;
        }
        else
          delete minf1;
      }
    }
    ftime( & t1 );
    info.tm = ( ( double )t1.time + ( double )t1.millitm / 1000 ) - ( ( double )t.time + ( double )t.millitm / 1000 );
    stop = true;
    if ( stop && mo->STOP_time ) stop = ( info.tm <= mo->STOP_maxtime );
    if ( stop && mo->STOP_iterations ) stop = ( info.cyc < mo->STOP_maxiterations );
    if ( stop && mo->STOP_optimization ) stop = ( minf->optval > mo->STOP_optimizationval );
    if ( stop && mo->STOP_improvement ) stop = ( info.cyc < lpass + lpass2 * mo->STOP_improvementint );
    if ( mo->DECOMP_enable ) CountAssignment( minf );
#if (mintrck > 0)
    fprintf( f, "%i %.3f %.3f %.3f %.3f %.3f %.3f %.3f %i %i %i %i %i\n", info.cyc, cd_time, exp_time, red_time, dom_time,
         cov_time, sprs_time, info.tm, buffers->minf->info.liters, buffers->minf->info.ON_terms,
         buffers->minf->info.groupterms, buffers->r_buffer->terms, buffers->minf->info.primes );
    fprintf( f1, "%i %.3f %.3f %.3f %.3f %.3f %.3f %.3f %i %i %i %i %i\n", info.cyc, cd_time, cd_time + exp_time,
         cd_time + exp_time + red_time, cd_time + exp_time + red_time + dom_time,
         cd_time + exp_time + red_time + dom_time + cov_time, cd_time + exp_time + red_time + dom_time + cov_time + sprs_time,
         info.tm, buffers->minf->info.liters, buffers->minf->info.ON_terms, buffers->minf->info.groupterms,
         buffers->r_buffer->terms, buffers->minf->info.primes );
#endif
#if (mintrck == 2)
    cd_time = 0;
    exp_time = 0;
    red_time = 0;
    dom_time = 0;
    cov_time = 0;
    sprs_time = 0;
    ftime( & tr_t );
#endif

    if ( mo->STOP_time == 1 ) i = gauge.GaugeStep( info.tm, 1 );
    else if ( mo->STOP_iterations == 1 ) i = gauge.GaugeStep( info.cyc, 1 );
    else
      i = gauge.GaugeStep( 0, 1 );
    if ( i != 0 ) stop = false;
  }
  while ( stop );
  if ( mo->COV_end == 1 )
  {
    if ( mo->DECOMP_COVenable == 1 ) minf1 = BlockCover( buffers->r_buffer );
    else if ( mo->COV_somin == 0 ) minf1 = Cover( buffers->r_buffer );
    else
      minf1 = SOCover( buffers->r_buffer );
    if ( mo->OR_no == 0 && srcf->outvars > 1 && mo->COV_somin == 0 ) SparseOutput( minf );
    minf->GetInfo();
  }

  delete buffers;
#if (mintrck > 0)
  fclose( f );
  fclose( f1 );
  fclose( f2 );
  fclose( f3 );
#endif
  gauge.CloseGauge();

  //IsolateFunction(minf);

  return minf;
}
//---------------------------------------------------------------------------
